# pylint: disable=wrong-import-position
# pylint: disable=protected-access

import base64
import io
import lzma
import os
import random
import shutil
import subprocess
import sys
from pathlib import Path
from typing import IO

import indexed_zstd
import pytest
import xz

try:
    # May not be installed with Python 3.14 because of incompatibilities.
    import zstandard
except ImportError:
    zstandard = None  # type: ignore

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from ratarmountcore.BlockParallelReaders import ParallelXZReader, ParallelZstdReader  # noqa: E402


def test_find_block():
    find_block = ParallelXZReader._find_block
    for i in range(-10, 10):
        assert find_block([], i) is None
        assert find_block([0], i) is None

    assert find_block([0, 1], -1) is None
    assert find_block([0, 1], 0) == 0
    assert find_block([0, 1], 1) is None

    assert find_block([0, 2], -1) is None
    assert find_block([0, 2], 0) == 0
    assert find_block([0, 2], 1) == 0
    assert find_block([0, 2], 2) is None

    assert find_block([0, 2, 4], -1) is None
    assert find_block([0, 2, 4], 0) == 0
    assert find_block([0, 2, 4], 1) == 0
    assert find_block([0, 2, 4], 2) == 1
    assert find_block([0, 2, 4], 3) == 1
    assert find_block([0, 2, 4], 4) is None


def test_block_size():
    blockSize = ParallelXZReader._block_size
    for i in range(-10, 10):
        assert blockSize([], i) == 0
        assert blockSize([0], i) == 0

    assert blockSize([0, 1], -1) == 0
    assert blockSize([0, 1], 0) == 1
    assert blockSize([0, 1], 1) == 0

    assert blockSize([0, 2], -1) == 0
    assert blockSize([0, 2], 0) == 2
    assert blockSize([0, 2], 1) == 0

    assert blockSize([0, 2, 4], -1) == 0
    assert blockSize([0, 2, 4], 0) == 2
    assert blockSize([0, 2, 4], 1) == 2
    assert blockSize([0, 2, 4], 2) == 0


@pytest.mark.parametrize("parallelization", [1, 2, 3, os.cpu_count()])
class TestParallelXZReader:
    @staticmethod
    def _create_archive(archivePath: Path, streams: int, blocksPerStream: int, blockSize: int):
        for _ in range(streams):
            tmpPath = str(archivePath) + '.tmp'
            size = blockSize * blocksPerStream
            Path(tmpPath).write_bytes(base64.b64encode(os.urandom(size))[:size])
            subprocess.run(['xz', f'--block-size={blockSize}', '--compress', '--force', tmpPath], check=True)
            shutil.copy(tmpPath + '.xz', archivePath)

    @staticmethod
    def _test_sequential_reading(archivePath: str, bufferSize: int, parallelization: int):
        with xz.open(archivePath, 'rb') as serialFile, (
            lzma.open(archivePath) if parallelization == 1 else ParallelXZReader(archivePath, parallelization)
        ) as parallelFile:
            bytesRead = 0
            while True:
                serialData = serialFile.read(bufferSize)
                parallelData = parallelFile.read(bufferSize)
                assert len(serialData) == len(parallelData)
                assert serialData == parallelData
                bytesRead += len(serialData)
                if len(serialData) < bufferSize:
                    break

            if hasattr(parallelFile, 'blockBoundaries'):
                assert bytesRead == parallelFile.blockBoundaries[-1]

    @staticmethod
    def _test_random_reads(archivePath: str, samples: int, parallelization: int):
        with xz.open(archivePath, 'rb') as serialFile, (
            lzma.open(archivePath) if parallelization == 1 else ParallelXZReader(archivePath, parallelization)
        ) as parallelFile:
            if hasattr(parallelFile, 'blockBoundaries'):
                size = parallelFile.blockBoundaries[-1]
            else:
                parallelFile.seek(io.SEEK_END)
                size = parallelFile.tell()
                parallelFile.seek(0)

            for _ in range(samples):
                offset = random.randint(0, size + 1)
                size = random.randint(0, (size + 1 - offset) * 2)  # half the time read past the end

                serialFile.seek(offset)
                serialData = serialFile.read(size)
                parallelFile.seek(offset)
                parallelData = parallelFile.read(size)
                assert len(serialData) == len(parallelData)
                assert serialData == parallelData

    def test_empty(self, parallelization, tmp_path):
        archivePath = tmp_path / 'test-archive.xz'
        self._create_archive(archivePath, 1, 1, 0)
        for bufferSize in [1, 2, 100, 1024, 128 * 1000]:
            self._test_sequential_reading(archivePath, bufferSize, parallelization)
        self._test_random_reads(archivePath, 10, parallelization)

    @pytest.mark.parametrize("streams", [1, 2, 7])
    @pytest.mark.parametrize("blocksPerStream", [1, 2, 7])
    @pytest.mark.parametrize("blockSize", [1, 2, 7])
    def test_small_multi_stream_block(self, parallelization, streams, blocksPerStream, blockSize, tmp_path):
        archivePath = os.path.join(str(tmp_path), 'test-archive.xz')
        self._create_archive(archivePath, streams, blocksPerStream, blockSize)
        for bufferSize in [1, 2, 100, 1024]:
            self._test_sequential_reading(archivePath, bufferSize, parallelization)
        self._test_random_reads(archivePath, 20, parallelization)

    @pytest.mark.parametrize("blockSize", [10, 333, 1024, 10 * 1000, 64 * 1024])
    def test_large_multi_stream_block(self, parallelization, blockSize, tmp_path):
        archivePath = os.path.join(str(tmp_path), 'test-archive.xz')
        self._create_archive(archivePath, 3, 4, blockSize)
        for bufferSize in [37, 1024, 128 * 1000]:
            self._test_sequential_reading(archivePath, bufferSize, parallelization)
        self._test_random_reads(archivePath, 200, parallelization)


class SeekableZstd:
    def __init__(self, filePath: str):
        self.name = filePath
        self.fileobj: IO[bytes] = io.BytesIO()
        self._rawFileObject = None
        self._reopen()

    def _reopen(self):
        if self.fileobj and not self.fileobj.closed:
            self.fileobj.close()
        if self._rawFileObject and not self._rawFileObject.closed:
            self._rawFileObject.close()

        self._rawFileObject = open(self.name, 'rb')
        self.fileobj = zstandard.ZstdDecompressor().stream_reader(self._rawFileObject, read_across_frames=True)

    def tell(self) -> int:
        return self.fileobj.tell()

    def read(self, size: int = -1) -> bytes:
        return self.fileobj.read(size)

    def close(self) -> None:
        return self.fileobj.close()

    def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
        assert whence == io.SEEK_SET
        if offset > self.fileobj.tell():
            self.read(offset - self.fileobj.tell())
        elif offset < self.fileobj.tell():
            self._reopen()
            self.fileobj.read(offset)
        return offset

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        if self.fileobj:
            self.fileobj.close()


@pytest.mark.parametrize("parallelization", [1, 2, 3, os.cpu_count()])
class TestParallelZstdReader:
    @staticmethod
    def _create_archive(archivePath: Path, streams: int, blocksPerStream: int, blockSize: int):
        for _ in range(streams):
            tmpPath = str(archivePath) + '.tmp'
            size = blockSize * blocksPerStream
            Path(tmpPath).write_bytes(base64.b64encode(os.urandom(size))[:size])
            subprocess.run(['zstd', '-q', f'--block-size={blockSize}', '--compress', '--force', tmpPath], check=True)
            shutil.copy(tmpPath + '.zst', archivePath)

    @staticmethod
    def _test_sequential_reading(archivePathLike: Path, bufferSize: int, parallelization: int):
        if zstandard is None:
            return

        archivePath = str(archivePathLike)

        with indexed_zstd.IndexedZstdFile(archivePath) as serialFile, (
            SeekableZstd(archivePath) if parallelization == 1 else ParallelZstdReader(archivePath, parallelization)
        ) as parallelFile:
            bytesRead = 0
            while True:
                serialData = serialFile.read(bufferSize)
                parallelData = parallelFile.read(bufferSize)
                assert len(serialData) == len(parallelData)
                assert serialData == parallelData
                bytesRead += len(serialData)
                if len(serialData) < bufferSize:
                    break

            if hasattr(parallelFile, 'blockBoundaries'):
                assert bytesRead == parallelFile.blockBoundaries[-1]

    @staticmethod
    def _test_random_reads(archivePathLike: Path, samples: int, parallelization: int):
        if zstandard is None:
            return

        archivePath = str(archivePathLike)

        with indexed_zstd.IndexedZstdFile(archivePath) as serialFile, (
            SeekableZstd(archivePath) if parallelization == 1 else ParallelZstdReader(archivePath, parallelization)
        ) as parallelFile:
            if hasattr(parallelFile, 'blockBoundaries'):
                size = parallelFile.blockBoundaries[-1]
            else:
                parallelFile.seek(io.SEEK_END)
                size = parallelFile.tell()
                if parallelization == 1:
                    parallelFile.close()
                    parallelFile = SeekableZstd(archivePath)
                else:
                    parallelFile.seek(0)

            for _ in range(samples):
                offset = random.randint(0, size + 1)
                size = random.randint(0, (size + 1 - offset) * 2)  # half the time read past the end

                serialFile.seek(offset)
                serialData = serialFile.read(size)

                # Files opened with the zstandard module cannot seek back not even in an emulated manner.
                if parallelization == 1 and offset < parallelFile.tell():
                    parallelFile.close()
                    parallelFile = SeekableZstd(archivePath)
                parallelFile.seek(offset)
                parallelData = parallelFile.read(size)

                assert len(serialData) == len(parallelData)
                assert serialData == parallelData

    def test_empty(self, parallelization, tmp_path):
        archivePath = tmp_path / 'test-archive.zst'
        self._create_archive(archivePath, 1, 1, 0)
        for bufferSize in [1, 2, 100, 1024, 128 * 1000]:
            self._test_sequential_reading(archivePath, bufferSize, parallelization)
        self._test_random_reads(archivePath, 10, parallelization)

    @pytest.mark.parametrize("streams", [1, 2, 7])
    @pytest.mark.parametrize("blocksPerStream", [1, 2, 7])
    @pytest.mark.parametrize("blockSize", [1, 2, 7])
    def test_small_multi_stream_block(self, parallelization, streams, blocksPerStream, blockSize, tmp_path):
        archivePath = tmp_path / 'test-archive.zst'
        self._create_archive(archivePath, streams, blocksPerStream, blockSize)
        for bufferSize in [1, 2, 100, 1024]:
            self._test_sequential_reading(archivePath, bufferSize, parallelization)
        self._test_random_reads(archivePath, 20, parallelization)

    @pytest.mark.parametrize("blockSize", [10, 333, 1024, 10 * 1000, 64 * 1024])
    def test_large_multi_stream_block(self, parallelization, blockSize, tmp_path):
        archivePath = tmp_path / 'test-archive.zst'
        self._create_archive(archivePath, 3, 4, blockSize)
        for bufferSize in [37, 1024, 128 * 1000]:
            self._test_sequential_reading(archivePath, bufferSize, parallelization)
        self._test_random_reads(archivePath, 200, parallelization)



import os
import stat
import time
from typing import IO, Any, Dict, Iterable, List, Optional, Set, Tuple, Union

from ratarmountcore.mountsource import FileInfo, MountSource, create_root_file_info, merge_statfs
from ratarmountcore.utils import overrides


class UnionMountSource(MountSource):
    def __init__(
        self,
        mountSources: List[MountSource],
        printDebug: int = 0,
        maxCacheDepth: int = 1024,
        maxCacheEntries: int = 100000,
        maxSecondsToCache: float = 60,
        # pylint: disable=unused-argument
        **options,
    ) -> None:
        """
        mountSources:
            List of mount sources for which to show a union view. The rightmost mount sources have
            the highest precedence. Meaning, if a file with the same name exists in multiple mount
            sources, then by default the file of the rightmost mount source will be returned.
        maxCacheEntries:
            Even assuming very long file paths like 1000 chars, the cache size
            will be below 100 MB if the maximum number of elements is 100k.
        maxSecondsToCache:
            Another problem is the setup time, as it might take ~0.001s for each lookup call
            and it shouldn't take minutes! Note that there always can be an edge case with hundred
            thousands of files in one folder, which can take an arbitrary amount of time to cache.
        """
        self.mountSources: List[MountSource] = mountSources
        self.printDebug = printDebug
        self.folderCache: Dict[str, List[MountSource]] = {"/": self.mountSources}
        self.folderCacheDepth = 0  # depth 1 means, we only cached top-level directories.
        self.rootFileInfo = create_root_file_info(userdata=[None])

        if len(self.mountSources) > 1:
            self._build_folder_cache(maxCacheDepth, maxCacheEntries, maxSecondsToCache)

    def _build_folder_cache(self, maxCacheDepth: int, maxCacheEntries: int, maxSecondsToCache: float) -> None:
        t0 = time.time()

        if self.printDebug >= 1:
            print(f"Building cache for union mount (timeout after {maxSecondsToCache}s)...")

        self.folderCache = {"/": [m for m in self.mountSources if m.is_immutable()]}

        lastFolderCache: Dict[str, List[MountSource]] = self.folderCache.copy()

        for depth in range(1, maxCacheDepth):
            # This intermediary structure is used because:
            #   1. We need to only iterate over the newly added folders in the next step
            #   2. We always want to (atomically) merge results for one folder depth so that we can be sure
            #      that if a folder of a cached depth can not be found in the cache that it does not exist at all.
            newFolderCache: Dict[str, List[MountSource]] = {}

            for folder, mountSources in lastFolderCache.items():
                for mountSource in mountSources:
                    filesInFolder = mountSource.list(folder)
                    if not filesInFolder:
                        continue

                    for file in filesInFolder:
                        if time.time() - t0 > maxSecondsToCache or maxCacheEntries <= 0:
                            return

                        fullPath = os.path.join(folder, file)
                        fileInfo = mountSource.lookup(fullPath)
                        if not fileInfo or not stat.S_ISDIR(fileInfo.mode):
                            continue

                        maxCacheEntries -= 1

                        if fullPath in newFolderCache:
                            newFolderCache[fullPath].append(mountSource)
                        else:
                            newFolderCache[fullPath] = [mountSource]

            if not newFolderCache:
                break

            self.folderCache.update(newFolderCache)
            self.folderCacheDepth = depth
            lastFolderCache = newFolderCache

        t1 = time.time()

        if self.printDebug >= 1:
            print(
                f"Cached mount sources for {len(self.folderCache)} folders up to a depth of "
                f"{self.folderCacheDepth} in {t1 - t0:.3}s for faster union mount."
            )

    @overrides(MountSource)
    def is_immutable(self) -> bool:
        return all(m.is_immutable() for m in self.mountSources)

    @overrides(MountSource)
    def lookup(self, path: str, fileVersion: int = 0) -> Optional[FileInfo]:
        if not path.strip('/'):
            return self.rootFileInfo.clone()

        if path in self.folderCache:
            # This case might be triggered when path is a folder
            cachedMountSources = self.folderCache[path]
        elif self.folderCache and self.folderCacheDepth > 0 and path.startswith('/'):
            # This should be the most common case, i.e., for regular files. Look up the parent folder in this case.
            parentFolder = '/'.join(path.split('/', self.folderCacheDepth + 1)[:-1])
            if not parentFolder:
                parentFolder = '/'
            cachedMountSources = self.folderCache.get(parentFolder, [])
        else:
            cachedMountSources = self.mountSources

        mountSources = [m for m in self.mountSources if not m.is_immutable() or m in cachedMountSources]

        # We need to keep the sign of the fileVersion in order to forward it to SQLiteIndexedTar.
        # When the requested version can't be found in a mount source, increment negative specified versions
        # by the amount of versions in that mount source or decrement the initially positive version.
        if fileVersion <= 0:
            for mountSource in reversed(mountSources):
                fileInfo = mountSource.lookup(path, fileVersion=fileVersion)
                if isinstance(fileInfo, FileInfo):
                    fileInfo.userdata.append(mountSource)
                    return fileInfo
                fileVersion += mountSource.versions(path)
                if fileVersion > 0:
                    break

        else:  # fileVersion >= 1
            for mountSource in mountSources:
                fileInfo = mountSource.lookup(path, fileVersion=fileVersion)
                if isinstance(fileInfo, FileInfo):
                    fileInfo.userdata.append(mountSource)
                    return fileInfo
                fileVersion -= mountSource.versions(path)
                if fileVersion < 1:
                    break

        return None

    @overrides(MountSource)
    def versions(self, path: str) -> int:
        return sum(mountSource.versions(path) for mountSource in self.mountSources)

    def _list(self, path: str, onlyMode: bool):
        files: Optional[Union[Set[str], Dict[str, FileInfo], Dict[str, int]]] = None

        for mountSource in reversed(self.mountSources):
            result = mountSource.list_mode(path) if onlyMode else mountSource.list(path)

            if files is None:
                if isinstance(result, dict):
                    files = result
                elif result is not None:
                    files = set(result)

            elif isinstance(result, dict):
                if isinstance(files, dict):
                    files.update(result)
                else:
                    files = files.union(result.keys())

            elif result is not None:
                # If one of the mount sources does not return extended information,
                # then strip it from all others and only return the names.
                if isinstance(files, dict):
                    files = set(files.keys())
                files = files.union(result)

        return files

    @overrides(MountSource)
    def list(self, path: str) -> Optional[Union[Iterable[str], Dict[str, FileInfo]]]:
        """
        Returns the set of all folder contents over all mount sources or None if the path was found in none of them.
        """
        return self._list(path, onlyMode=False)

    @overrides(MountSource)
    def list_mode(self, path: str) -> Optional[Union[Iterable[str], Dict[str, int]]]:
        """
        Returns the set of all folder contents over all mount sources or None if the path was found in none of them.
        """
        return self._list(path, onlyMode=True)

    @overrides(MountSource)
    def open(self, fileInfo: FileInfo, buffering=-1) -> IO[bytes]:
        mountSource = fileInfo.userdata.pop()
        try:
            assert isinstance(mountSource, MountSource)
            return mountSource.open(fileInfo, buffering=buffering)
        finally:
            fileInfo.userdata.append(mountSource)

    @overrides(MountSource)
    def read(self, fileInfo: FileInfo, size: int, offset: int) -> bytes:
        mountSource = fileInfo.userdata.pop()
        try:
            assert isinstance(mountSource, MountSource)
            return mountSource.read(fileInfo, size, offset)
        finally:
            fileInfo.userdata.append(mountSource)

    @overrides(MountSource)
    def list_xattr(self, fileInfo: FileInfo) -> List[str]:
        mountSource = fileInfo.userdata.pop()
        try:
            return mountSource.list_xattr(fileInfo) if isinstance(mountSource, MountSource) else []
        finally:
            fileInfo.userdata.append(mountSource)

    @overrides(MountSource)
    def get_xattr(self, fileInfo: FileInfo, key: str) -> Optional[bytes]:
        mountSource = fileInfo.userdata.pop()
        try:
            return mountSource.get_xattr(fileInfo, key) if isinstance(mountSource, MountSource) else None
        finally:
            fileInfo.userdata.append(mountSource)

    @overrides(MountSource)
    def get_mount_source(self, fileInfo: FileInfo) -> Tuple[str, MountSource, FileInfo]:
        sourceFileInfo = fileInfo.clone()
        mountSource = sourceFileInfo.userdata.pop()

        if not isinstance(mountSource, MountSource):
            return '/', self, fileInfo

        # Because all mount sources are mounted at '/', we do not have to append
        # the mount point path returned by get_mount_source to the mount point '/'.
        return mountSource.get_mount_source(sourceFileInfo)

    @overrides(MountSource)
    def statfs(self) -> Dict[str, Any]:
        return merge_statfs([mountSource.statfs() for mountSource in self.mountSources], printDebug=self.printDebug)

    @overrides(MountSource)
    def __exit__(self, exception_type, exception_value, exception_traceback):
        for mountSource in self.mountSources:
            mountSource.__exit__(exception_type, exception_value, exception_traceback)

    def join_threads(self):
        for mountSource in self.mountSources:
            if hasattr(mountSource, 'join_threads'):
                mountSource.join_threads()



# pylint: disable=wrong-import-position
# pylint: disable=redefined-outer-name

import io
import os
import stat
import sys

import pytest

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from ratarmountcore.mountsource.compositing.singlefile import SingleFileMountSource  # noqa: E402


class TestSingleFileMountSource:
    @staticmethod
    @pytest.mark.parametrize('path', ["foo", "/foo", "folder/foo", "/folder/./foo", "/a/b/c/foo"])
    def test_single_file(path: str):
        contents = b"bar"
        ms = SingleFileMountSource(path, io.BytesIO(contents))

        # The mount source API expects normalized paths. Only the one-time constructor is lenient enough to
        # accept non-normalized paths.
        path = os.path.normpath('/' + path).lstrip('/')

        fileInfo = ms.lookup(path)
        assert fileInfo
        assert stat.S_ISREG(fileInfo.mode)
        assert not stat.S_ISDIR(fileInfo.mode)
        assert fileInfo.size == len(contents)

        with ms.open(fileInfo) as file:
            assert file.read() == contents
        assert ms.read(fileInfo, size=len(contents) + 1, offset=0) == contents

        splitPath = path.split('/')
        for i in range(len(splitPath)):
            subpath = '/'.join(['', *splitPath[:i]])
            for queryPath in (subpath, subpath.lstrip('/')):
                fileInfo = ms.lookup(queryPath)
                assert fileInfo
                assert stat.S_ISDIR(fileInfo.mode)
                assert not stat.S_ISREG(fileInfo.mode)

                files_mode = ms.list_mode(queryPath)
                assert files_mode
                assert len(list(files_mode)) == 1

                files = ms.list(queryPath)
                assert files
                assert len(list(files)) == 1



# pylint: disable=wrong-import-order
# pylint: disable=wrong-import-position
# pylint: disable=protected-access

import io
import os
import sys
import tarfile
import tempfile
from pathlib import Path

import fsspec

try:
    import pandas as pd
except ImportError:
    pd = None  # type: ignore


sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from ratarmountcore.SQLiteIndexedTarFsspec import SQLiteIndexedTarFileSystem as RatarFS  # noqa: E402


def find_test_file(relativePathOrName):
    for i in range(3):
        path = os.path.sep.join([".."] * i + ["tests", relativePathOrName])
        if os.path.exists(path):
            return path
    return relativePathOrName


def test_file_system():
    fs = RatarFS(find_test_file('single-file.tar'))

    assert 'bar' in fs.ls("/", detail=False)
    assert 'bar' in [info['name'] for info in fs.ls("/", detail=True)]

    assert not fs.isfile("/")
    assert fs.isdir("/")
    assert fs.exists("/")

    assert fs.isfile("/bar")
    assert not fs.isdir("/bar")
    assert not fs.exists("/bar2")

    assert fs.cat("/bar") == b"foo\n"
    assert fs.cat("bar") == b"foo\n"

    with fs.open("bar") as file:
        assert file.read() == b"foo\n"


def test_url_context_manager():
    with fsspec.open("ratar://bar::file://" + find_test_file('single-file.tar')) as file:
        assert file.read() == b"foo\n"


def test_url():
    openedFile = fsspec.open("ratar://bar::file://" + find_test_file('single-file.tar'))
    with openedFile as file:
        assert file.read() == b"foo\n"


def test_pandas():
    if pd is None:
        return

    with tempfile.TemporaryDirectory(suffix=".test.ratarmount") as folderPath:
        oldPath = os.getcwd()
        os.chdir(folderPath)
        try:
            Path("test.csv").write_bytes(b"1,2\n3,4")
            with tarfile.open("test-csv.tar", "w") as archive:
                archive.add("test.csv")

            # Pandas seems
            data = pd.read_csv("ratar://test.csv::file://test-csv.tar", compression=None, header=None)
            assert data.iloc[0, 1] == 2
        finally:
            os.chdir(oldPath)


if False:
    # I had problems with resource deallocation!
    # For Rapidgzip it becomes important because of the background threads.
    # I can only reproduce this bug when run in global namespace.
    # It always works without problems inside a function.
    # TODO I don't know how to fix this. Closing the file object in SQLiteIndexedTar.__del__
    #      would fix this particular error, but it would lead to other errors for recursive mounting
    #      and when using fsspec.open chained URLs...
    #      Only calling join_threads also does not work for some reason.
    #      Checking with sys.getrefcount and only closing it if it is the only one left also does not work
    #      because the refcount is 3 inside __del__ for some reason.
    #      Closing the file inside RapidgzipFile.__del__ also does not work because it results in the
    #      same error during that close call, i.e., it is already too late at that point. I'm not sure
    #      why it is too late there but not too late during the SQLiteIndexedTar destructor...
    #      Maybe there are also some cyclic dependencies?
    with tempfile.TemporaryDirectory(suffix=".test.ratarmount") as folderPath2:
        contents = os.urandom(96 * 1024 * 1024)

        tarPath = os.path.join(folderPath2, "random-data.tar.gz")
        with tarfile.open(name=tarPath, mode="w:gz") as tarArchive:
            # Must create a sufficiently large .tar.gz so that rapidgzip is actually used.
            # In the future this "has multiple chunks" rapidgzip test is to be removed and
            # this whole test becomes redundant.
            tinfo = tarfile.TarInfo("random-data")
            tinfo.size = len(contents)
            tarArchive.addfile(tinfo, io.BytesIO(contents))

        # Only global variables trigger the "Detected Python finalization from running rapidgzip thread." bug.
        # I am not sure why. Probably, because it gets garbage-collected later.
        globalOpenFile = fsspec.open("ratar://random-data::file://" + tarPath)
        with globalOpenFile as openedFile2:
            assert openedFile2.read() == contents

        # This is still some step the user has to do, but it cannot be avoided.
        # It might be helpful if fsspec had some kind of better resource management for filesystems though.
        del globalOpenFile



# pylint: disable=unused-import
# pylint: disable=wrong-import-position
# pylint: disable=protected-access

import os
import sys

from helpers import find_test_file

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

# Appends proper format checkers to 7Z, EXT4, FAT, RAR formats.
import ratarmountcore.mountsource.archives  # noqa: E402, F401
from ratarmountcore.formats import (  # noqa: E402
    COMPRESSION_FORMATS,
    FILE_FORMATS,
    FileFormatID,
    detect_formats,
    might_be_format,
)


def test_format_detection():
    # This test assumes that we use correct extensions for all files in the tests folder.
    folder = os.path.dirname(find_test_file("tests/single-file.tar"))
    for name in os.listdir(folder):
        path = os.path.join(folder, name)
        if not os.path.isfile(path):
            continue

        with open(path, 'rb') as file:
            # The caching should not change the results!
            formats = detect_formats(file)
            assert formats == {fid for fid, info in FILE_FORMATS.items() if might_be_format(file, info)}, name
            assert formats == {fid for fid in FILE_FORMATS if might_be_format(file, fid)}, name

            splitName = name.rsplit('.', 1)
            if len(splitName) > 1 and name and name[0] != '.':
                extension = splitName[-1]
                message = f"{name} with extension {extension} was detected as possibly {formats}"
                # Except for encrypted SQLAR, the format and extension should match.
                # The only chimera file: chimera-tbz2-zip has explicitly no extension!
                if extension != 'sqlar' and FileFormatID.SQLAR in formats:
                    formats.remove(FileFormatID.SQLAR)

                # The name should be self-explanatory. Do not test for it being recognized as ZIP
                # because it is not a bug if it is not recognized as such.
                if name == 'rar-misrecognized-as-zip.rar' and FileFormatID.ZIP in formats:
                    formats.remove(FileFormatID.ZIP)

                # The file is still large enough that it should be recognized as a (truncated) TAR.
                if name == 'single-file-split.tar.001':
                    extension = 'tar'

                if name == 'single-file-split.tar.002':
                    # This and empty.tar are both filled only with zero-bytes!
                    pass
                elif len(formats) == 0:
                    assert extension in ['001', '002', 'ini', 'sh', 'snar', 'txt', 'py'], message
                elif len(formats) == 1:
                    formatID = next(iter(formats))
                    extensions = FILE_FORMATS[formatID].extensions
                    if formatID in COMPRESSION_FORMATS:
                        extensions += ['t' + e for e in extensions]
                    if formatID == FileFormatID.RATARMOUNT_INDEX:
                        extensions.append('sqlite')
                    assert extension in extensions, message
                elif len(formats) > 1:
                    # SQLite files can be Ratarmount indexes or SQLAR
                    # EXT4 images can be interpreted as TAR files (with only zero blocks at the start).
                    assert extension in ['ext4', 'sqlar'], message



# pylint: disable=wrong-import-position
# pylint: disable=protected-access

import os
import stat
import sys

from helpers import copy_test_file

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import pytest  # noqa: E402
from ratarmountcore.mountsource.compositing.automount import AutoMountLayer  # noqa: E402
from ratarmountcore.mountsource.factory import open_mount_source  # noqa: E402


# @pytest.mark.parametrize("parallelization", [1, 2, 4])
@pytest.mark.parametrize("parallelization", [1])
class TestAutoMountLayer:
    @staticmethod
    def test_regex_mount_point_tar(parallelization):
        options = {
            'clearIndexCache': True,
            'recursive': True,
            'parallelization': parallelization,
            'transformRecursiveMountPoint': ('.*/([^/]*).tar', r'\1'),
        }

        with copy_test_file("packed-100-times.tar.gz") as path, open_mount_source(path, **options) as mountSource:
            recursivelyMounted = AutoMountLayer(mountSource, **options)

            assert recursivelyMounted.list('/')
            assert recursivelyMounted.list('/ufo_12')
            assert recursivelyMounted.list('/ufo_00')
            assert list(recursivelyMounted.list('/ufo_00').keys()) == ['ufo']
            assert recursivelyMounted.open(recursivelyMounted.lookup('/ufo_00/ufo')).read() == b'iriya\n'

    @staticmethod
    def test_regex_mount_point_tar_gz(parallelization):
        options = {
            'clearIndexCache': True,
            'recursive': True,
            'parallelization': parallelization,
            'transformRecursiveMountPoint': ('.*/([^/]*).tar.gz', r'\1'),
        }

        # TODO Using the compressed-1000-times.tar.gz, which is ~200 KiB compressed and uncompressed leads to 12 GiB
        #      of memory usage!! And not only that, after this, that memory is not freed for the subsequent tests in
        #      other files and those other files will actually take 10x or more longer than without this test running
        #      before! It might be that the memory usage makes Python's garbage collector a bottleneck because of too
        #      many small objects?!
        with copy_test_file("compressed-100-times.tar.gz") as path, open_mount_source(path, **options) as mountSource:
            recursivelyMounted = AutoMountLayer(mountSource, **options)

            assert recursivelyMounted.list('/')
            assert recursivelyMounted.list('/ufo_12')
            assert recursivelyMounted.list('/ufo_00')
            assert list(recursivelyMounted.list('/ufo_00').keys()) == ['ufo']
            assert recursivelyMounted.open(recursivelyMounted.lookup('/ufo_00/ufo')).read() == b'iriya\n'

    @staticmethod
    def test_regex_mount_point_gz(parallelization):
        options = {
            'clearIndexCache': True,
            'recursive': True,
            'parallelization': parallelization,
            'transformRecursiveMountPoint': ('.*/([^/]*).gz', r'\1'),
        }

        # For some reason, the test with 1000 recursion fails reproducibly after ~196 depth, therefore use only 100.
        # > Recursively mounted: /ufo_805.gz
        # >  File "core/ratarmountcore/SQLiteIndexedTar.py", line 2085, in _detectTar
        # > indexed_gzip.indexed_gzip.ZranError: zran_read returned error: ZRAN_READ_FAIL (file: n/a)
        # On MacOS, the test fails even sooner after only 7 layers:
        #     core/tests/test_AutoMountLayer.py::TestAutoMountLayer::test_regex_mount_point_gz[1]
        #     [Info] Try to open with rarfile
        #     [Info] Try to open with tarfile
        #     [Info] Detected compression gz for file object: <_io.BufferedReader
        #         name='/var/folders/24/8k4/T/tmp5/compressed-100-times.gz'>
        #     [Info] Do not reopen with rapidgzip backend because:
        #     [Info]  - is too small to qualify for parallel decompression.
        #     [Info] Undid gz file compression by using: IndexedGzipFile
        #     [...]
        #     Recursively mounted: /ufo_94
        #     [Info] Try to open with rarfile
        #     [Info] Try to open with tarfile
        #     [Info] Detected compression gz for file object: <StenciledFile>
        #     [Info] Do not reopen with rapidgzip backend because:
        #     [Info]  - the file to open is a recursive file, which limits the usability of
        #     [Info]    parallel decompression.
        #     [Info]  - is too small to qualify for parallel decompression.
        #     [Info] Undid gz file compression by using: IndexedGzipFile
        #     [Info] File object <IndexedGzipFile> is not a TAR.
        #     Creating offset dictionary for ufo_93.gz ...
        #     Did not find any file in the given TAR: ufo_93.gz. Assuming a compressed file.
        #     Resorting files by path ...
        #     Creating offset dictionary for ufo_93.gz took 0.00s
        #     [Info] Tried to write the database to disk but found no other path than: :memory:
        #     [Info] Will skip storing compression seek data because the database is in memory.
        #     [Info] If the database is in memory, then this data will not be read anyway.
        #     /Users/runner/work/_temp/b97a4216-0bf9-484c-b63a-20f214563ba6.sh: line 3: 6181
        #         Illegal instruction: 4  pytest -s --verbosity=3 --disable-warnings "$file"
        #     [Info] Opened archive with tarfile backend.
        #     Error: Process completed with exit code 132.
        # Probably a literal stack overflow because of the recursion, which results in a function return to
        # garbage data resulting in an illegal instruction?
        if sys.platform.startswith('darwin'):
            return

        with copy_test_file("compressed-100-times.gz") as path, open_mount_source(path, **options) as mountSource:
            recursivelyMounted = AutoMountLayer(mountSource, **options)

            assert recursivelyMounted.list('/')
            assert recursivelyMounted.list('/ufo_12')
            assert recursivelyMounted.list('/ufo_00')
            fileInfo = recursivelyMounted.lookup('/ufo_00/ufo')
            assert recursivelyMounted.open(fileInfo).read() == b'iriya\n'

    @staticmethod
    def test_file_versions(parallelization):
        options = {
            'clearIndexCache': True,
            'recursive': True,
            'parallelization': parallelization,
        }

        with copy_test_file("tests/double-compressed-nested-tar.tgz.tgz") as path, open_mount_source(
            path, **options
        ) as mountSource:
            recursivelyMounted = AutoMountLayer(mountSource, **options)

            for folder in ['/', '/nested-tar.tar.gz', '/nested-tar.tar.gz/foo', '/nested-tar.tar.gz/foo/fighter']:
                assert recursivelyMounted.lookup(folder)
                assert recursivelyMounted.list(folder)
                assert recursivelyMounted.versions(folder) > 0

            for mountedFile in ['/nested-tar.tar.gz']:
                assert recursivelyMounted.versions(folder) > 0
                assert stat.S_ISREG(recursivelyMounted.lookup(mountedFile, fileVersion=1).mode)

            # assert recursivelyMounted.open(recursivelyMounted.lookup('/ufo_00/ufo')).read() == b'iriya\n'

    @staticmethod
    @pytest.mark.parametrize("recursive", [False])
    @pytest.mark.parametrize("maxRecursionDepth", range(7))
    def test_recursion_depth(parallelization, recursive, maxRecursionDepth):
        name = "triple-compressed-nested-tar"
        archivePath = f"tests/{name}.tgz.tgz.gz"
        options = {
            'clearIndexCache': True,
            'recursive': recursive,
            'recursionDepth': maxRecursionDepth,
            'parallelization': parallelization,
        }
        with copy_test_file(archivePath) as path, open_mount_source(path, **options) as mountSource:
            recursivelyMounted = AutoMountLayer(mountSource, **options)
            assert recursivelyMounted.list('/')

            maxDepth = 6
            recursionDepth = maxDepth if maxRecursionDepth is None else maxRecursionDepth

            def check_directory(fileInfo):
                assert fileInfo
                assert stat.S_ISDIR(fileInfo.mode)

            def check_file(fileInfo, size=None):
                assert fileInfo
                assert fileInfo.size == size
                assert not stat.S_ISDIR(fileInfo.mode)
                assert stat.S_ISREG(fileInfo.mode)

            def check_non_existing(fileInfo):
                assert not fileInfo

            # This was insane brainfuck... Off by one errors everywhere.
            # Check with: ratarmount -f -c --recursion-depth 1 tests/triple-compressed-nested-tar.tgz.tgz.gz mounted
            nestedRoot = f"/{name}.tgz.tgz/nested-tar.tar.gz"
            testPaths = {
                f"/{name}.tgz.tgz": {
                    0: lambda path: check_file(path, size=436),
                    1: lambda path: check_directory,
                },
                f"/{name}.tgz.tgz/{name}.tgz.tgz": {
                    1: lambda path: check_file(path, size=10240),
                    2: lambda path: check_directory,
                },
                nestedRoot: {
                    2: lambda path: check_file(path, size=291),
                    3: lambda path: check_directory,
                },
                # This file is only visible when only the gzip compression has been undone.
                # And the name is probably the one stored in gzip itself. It has support for that!
                # For the next recursion, this virtual file will be replaced by the actual contents.
                nestedRoot
                + "/nested-tar.tar": {
                    3: lambda path: check_file(path, size=20480),
                    4: check_non_existing,
                },
                nestedRoot
                + "/foo": {
                    4: lambda path: check_directory,
                },
                nestedRoot
                + "/foo/fighter": {
                    4: lambda path: check_directory,
                },
                nestedRoot
                + "/foo/fighter/ufo": {
                    4: lambda path: check_file(path, size=6),
                },
                nestedRoot
                + "/foo/lighter.tar": {
                    4: lambda path: check_file(path, size=10240),
                    5: lambda path: check_directory,
                },
                nestedRoot
                + "/foo/lighter.tar/fighter": {
                    5: lambda path: check_directory,
                },
                nestedRoot
                + "/foo/lighter.tar/fighter/bar": {
                    5: lambda path: check_file(path, size=10),
                },
            }

            # Check paths up to the allowed recursion depth
            for path, configurations in testPaths.items():
                fileInfo = recursivelyMounted.lookup(path)

                # Files should not exist when their actual depth is larger than the specified mount depth.
                if recursionDepth < min(configurations.keys()):
                    assert not recursivelyMounted.exists(path)
                    assert fileInfo is None
                    continue

                expectedDepth = recursionDepth if recursionDepth in configurations else max(configurations.keys())
                configurations[expectedDepth](fileInfo)



# pylint: disable=protected-access, import-outside-toplevel, unused-argument

import contextlib
import io
import json
import os
import re
import stat
import tarfile
import zlib
from timeit import default_timer as timer
from typing import IO, Any, Dict, List, Optional, Tuple, Union

try:
    import deflate
except ImportError:
    deflate = None

try:
    from isal import isal_zlib
except ImportError:
    isal_zlib = None  # type: ignore

try:
    import PySquashfsImage
except ImportError:
    PySquashfsImage = None  # type: ignore

try:
    from PySquashfsImage import SquashFsImage
except ImportError:
    # We need to define this for @overrides and pytype, but it also is a nice documentation
    # for the expected members in PySquashfsImage.SquashFsImage.
    class SquashFsImage:  # type: ignore
        def __init__(self, fd, offset: int = 0, closefd: bool = True) -> None:
            self._sblk: Any = None

        def __iter__(self):
            pass

        def _get_compressor(self, compression_id):
            pass

        def _initialize(self):
            pass

        # These are not overridden, only called:

        def _read_block_list(self, start, offset, blocks):
            raise NotImplementedError

        def _read_fragment(self, fragment):
            raise NotImplementedError

        def _read_inode(self, start_block, offset):
            raise NotImplementedError

        def _opendir(self, block_start, offset):
            raise NotImplementedError

        def _dir_scan(self, start_block, offset):
            raise NotImplementedError


try:
    from PySquashfsImage.compressor import Compression, Compressor, compressors
except ImportError:
    Compressor = object

from ratarmountcore.formats import find_squashfs_offset
from ratarmountcore.mountsource import FileInfo, MountSource
from ratarmountcore.mountsource.SQLiteIndexMountSource import SQLiteIndexMountSource
from ratarmountcore.SQLiteIndex import SQLiteIndex, SQLiteIndexedTarUserData
from ratarmountcore.utils import overrides


class IsalZlibDecompressor(Compressor):
    name = "gzip"

    def __init__(self):
        self._decompress = zlib.decompress if isal_zlib is None else isal_zlib.decompress

    def uncompress(self, src, size, outsize):
        return self._decompress(src)


class LibdeflateZlibDecompressor(Compressor):
    name = "gzip"

    def __init__(self):
        self._lib = deflate

    def uncompress(self, src, size, outsize):
        # Beware: https://github.com/dcwatson/deflate/issues/41
        return self._lib.zlib_decompress(src, outsize)


class LZ4Compressor(Compressor):
    name = "lz4"

    def __init__(self):
        import lz4.block

        self._lib = lz4.block

    def uncompress(self, src, size, outsize):
        return self._lib.decompress(src, outsize)


class LZMACompressor(Compressor):
    name = "lzma"

    def __init__(self, blockSize):
        self._block_size = blockSize
        try:
            import lzma
        except ImportError:
            from backports import lzma
        self._lib = lzma

    def uncompress(self, src, size, outsize):
        # https://github.com/plougher/squashfs-tools/blob/a04910367d64a5220f623944e15be282647d77ba/squashfs-tools/
        #   lzma_wrapper.c#L40
        # res = LzmaCompress(dest + LZMA_HEADER_SIZE, &outlen, src, size, dest,
        #                    &props_size, 5, block_size, 3, 0, 2, 32, 1);
        # https://github.com/jljusten/LZMA-SDK/blob/781863cdf592da3e97420f50de5dac056ad352a5/C/LzmaLib.h#L96
        # -> level=5, dictSize=block_size, lc=3, lp=0, pb=2, fb=32, numThreads=1
        # https://github.com/plougher/squashfs-tools/blob/a04910367d64a5220f623944e15be282647d77ba/squashfs-tools/
        #   lzma_wrapper.c#L30
        # For some reason, squashfs does not store raw lzma but adds a custom header of 5 B and 8 B little-endian
        # uncompressed size, which can be read with struct.unpack('<Q', src[5:5+8]))
        LZMA_PROPS_SIZE = 5
        LZMA_HEADER_SIZE = LZMA_PROPS_SIZE + 8
        return self._lib.decompress(
            src[LZMA_HEADER_SIZE:],
            format=self._lib.FORMAT_RAW,
            filters=[{"id": self._lib.FILTER_LZMA1, 'lc': 3, 'lp': 0, 'pb': 2, 'dict_size': self._block_size}],
        )


class SquashFSFile(io.RawIOBase):
    def __init__(self, image, inode) -> None:
        self._image = image
        self._inode = inode

        self._offset = 0
        self._size = inode.data
        self._block_size = image._sblk.block_size
        self._lastBlockIndex = inode.data // self._block_size

        self._blockList = []
        self._dataToBlockOffset: Dict[int, int] = {}  # block offset may be negative (-size) for sparse blocks
        self._compressedBlockOffsets = []
        if inode.blocks:
            self._blockList = [
                block
                for block in image._read_block_list(inode.block_start, inode.block_offset, inode.blocks)
                if block != PySquashfsImage.SQUASHFS_INVALID_FRAG
            ]

            compressedBlockOffset = inode.start
            for i, block in enumerate(self._blockList):
                blockSize = self._size % self._block_size if i == self._lastBlockIndex else self._block_size
                assert blockSize > 0
                if block:
                    self._compressedBlockOffsets.append(compressedBlockOffset)
                    compressedBlockOffset += PySquashfsImage.SQUASHFS_COMPRESSED_SIZE_BLOCK(block)
                else:
                    # sparse file
                    self._compressedBlockOffsets.append(-blockSize)
            assert len(self._compressedBlockOffsets) == len(self._blockList)

        self._fragment = None
        if inode.frag_bytes:
            self._fragment = image._read_fragment(inode.fragment)

        self._bufferIO: Optional[IO[bytes]] = None
        self._blockIndex = 0
        self._buffer = b''
        self._refill_buffer(self._blockIndex)  # Requires self._blockList to be initialized

    def _refill_buffer(self, blockIndex: int) -> None:
        self._blockIndex = blockIndex
        self._buffer = b''

        assert self._blockIndex >= 0
        if self._blockIndex < len(self._blockList):
            block = self._blockList[self._blockIndex]
            if block:
                start = self._compressedBlockOffsets[self._blockIndex]
                self._buffer = self._image._read_data_block(start, block)
            else:
                if (self._blockIndex + 1) * self._block_size >= self._size:
                    blockSize = max(0, self._size - self._blockIndex * self._block_size)
                else:
                    blockSize = self._block_size
                self._buffer = b'\0' * blockSize
        elif self._fragment and self._blockIndex == len(self._blockList):
            fragment = self._image._read_data_block(*self._fragment)
            self._buffer = fragment[self._inode.offset : self._inode.offset + self._inode.frag_bytes]

        self._bufferIO = io.BytesIO(self._buffer)

    @overrides(io.RawIOBase)
    def readinto(self, buffer):
        """Generic implementation which uses read."""
        with memoryview(buffer) as view, view.cast("B") as byteView:  # type: ignore
            readBytes = self.read(len(byteView))
            byteView[: len(readBytes)] = readBytes
        return len(readBytes)

    def read1(self, size: int = -1) -> bytes:
        if not self._bufferIO:
            raise RuntimeError("Closed file cannot be read from!")
        result = self._bufferIO.read(size)
        # An empty buffer signals the end of the file!
        if result or not self._buffer:
            return result

        self._blockIndex += 1
        self._refill_buffer(self._blockIndex)
        return self._bufferIO.read(size)

    @overrides(io.RawIOBase)
    def read(self, size: int = -1) -> bytes:
        result = bytearray()
        while size < 0 or len(result) < size:
            read_data = self.read1(size if size < 0 else size - len(result))
            if not read_data:
                break
            result.extend(read_data)
        return bytes(result)

    @overrides(io.RawIOBase)
    def fileno(self) -> int:
        # This is a virtual Python level file object and therefore does not have a valid OS file descriptor!
        raise io.UnsupportedOperation

    @overrides(io.RawIOBase)
    def seekable(self) -> bool:
        return True

    @overrides(io.RawIOBase)
    def readable(self) -> bool:
        return True

    @overrides(io.RawIOBase)
    def writable(self) -> bool:
        return False

    @overrides(io.RawIOBase)
    def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
        if not self._bufferIO:
            raise RuntimeError("Closed file cannot be seeked!")

        here = self.tell()
        if whence == io.SEEK_CUR:
            offset += here
        elif whence == io.SEEK_END:
            offset += self._size

        self._offset = max(0, min(offset, self._size))
        bufferOffset = self._blockIndex * self._block_size
        if offset < bufferOffset or offset >= bufferOffset + len(self._buffer):
            self._refill_buffer(offset // self._block_size)  # Updates self._blockIndex!
        self._bufferIO.seek(offset - self._blockIndex * self._block_size)

        return self.tell()

    @overrides(io.RawIOBase)
    def tell(self) -> int:
        # Returning self._blockIndex * self._block_size + self._bufferIO.tell() will not work when we have
        # an empty buffer after trying to read past the end of the file.
        return self._offset


# https://github.com/matteomattei/PySquashfsImage/blob/e637b26b3bc6268dd589fa1439fecf99e49a565b/PySquashfsImage/__init__.py#L82
class SquashFSImage(SquashFsImage):
    """
    Contains several improvements over the base class:
     - Does not create the whole folder hierarchy in memory when only iterating over it to avoid high memory
       usage for SquashFS images with millions of files.
     - Adds seekable, streamable file object accessor that can be opened given a single number.
     - Adds thread locks around the underlying file object so that multiple file objects can be opened and used
       from multiple threads concurrently.
     - Uses libdeflate or ISA-L if installed, which a generally faster than the standard zlib.
     - Fixes lz4 support. (Merged into PySquashfsImage upstream, but not released yet.)
     - Adds lzma support. (Merged into PySquashfsImage upstream, but not released yet.)

    Beware that we are overwriting and using "private" methods starting with underscores!
    That's why we need to pin to an exact PySquashfsImage release.
    """

    @overrides(SquashFsImage)
    def __init__(self, *args, **kwargs):
        self._real_root = None
        super().__init__(*args, **kwargs)  # Calls overridden _initialize

    @overrides(SquashFsImage)
    def _get_compressor(self, compression_id):
        if compression_id == Compression.ZLIB:
            if deflate is not None:
                return LibdeflateZlibDecompressor()
            if isal_zlib is not None:
                return IsalZlibDecompressor()
        if compression_id == Compression.LZ4:
            return LZ4Compressor()
        if compression_id == Compression.LZMA:
            return LZMACompressor(self._sblk.block_size)
        if compression_id not in compressors:
            raise ValueError("Unknown compression method " + compression_id)
        return compressors[compression_id]()

    @overrides(SquashFsImage)
    def _initialize(self):
        self._fd.seek(self._offset)
        self._read_super()
        self._read_uids_guids()
        self._read_fragment_table()
        self._read_xattrs_from_disk()
        # Moved self._root initialization into a property and _generate_root

    def _generate_root(self):
        root_block = PySquashfsImage.SQUASHFS_INODE_BLK(self._sblk.root_inode)
        root_offset = PySquashfsImage.SQUASHFS_INODE_OFFSET(self._sblk.root_inode)
        self._real_root = self._dir_scan(root_block, root_offset)

    @staticmethod
    def _join_inode_offset(start_block, offset):
        assert start_block < 2**32
        assert offset < 2**16
        return (start_block << 16) + offset

    @staticmethod
    def _split_inode_offset(inode_offset):
        return inode_offset >> 16, inode_offset & 0xFFFF

    def read_inode(self, inode_offset):
        """Newly added function over SquashFsImage that adds an accessor via a simple integer."""
        return self._read_inode(*self._split_inode_offset(inode_offset))

    @overrides(SquashFsImage)
    def __iter__(self):  # -> PySquashfsImage.file.File
        """
        Performance improved function over PySquashfsImage.__iter__ that generates data on demand instead
        of keeping all metadata in memory and returning a generator over that.
        """
        root_block = PySquashfsImage.SQUASHFS_INODE_BLK(self._sblk.root_inode)
        root_offset = PySquashfsImage.SQUASHFS_INODE_OFFSET(self._sblk.root_inode)
        root_inode_offset, root_directory = self._open_directory(root_block, root_offset)
        yield root_inode_offset, root_directory
        yield from self._recursive_inodes_iterator(root_directory)

    def _open_directory(self, start_block, offset, parent=None, name=None):
        directory = self._opendir(start_block, offset)
        if parent is not None:
            directory._parent = parent
        if name is not None:
            directory._name = name
        return self._join_inode_offset(start_block, offset), directory

    def _recursive_inodes_iterator(self, directory):  # -> PySquashfsImage.file.File
        for entry in directory.entries:
            start_block = entry["start_block"]
            offset = entry["offset"]
            if entry["type"] == PySquashfsImage.Type.DIR:
                inode_offset, subdirectory = self._open_directory(start_block, offset, directory, entry["name"])
                yield inode_offset, subdirectory
                yield from self._recursive_inodes_iterator(subdirectory)
            else:
                inode = self._read_inode(start_block, offset)
                cls = PySquashfsImage.filetype[entry["type"]]
                yield self._join_inode_offset(start_block, offset), cls(self, inode, entry["name"], directory)

    @property
    def _root(self):
        if self._real_root is None:
            self._generate_root()
        return self._real_root

    @_root.setter
    def _root(self, value):
        # super().__init__ will initialize it to None but super()._initialize should not be called!
        assert value is None

    def open(self, inode):
        return SquashFSFile(self, inode)


class SquashFSMountSource(SQLiteIndexMountSource):
    # fmt: off
    def __init__(
        self,
        fileOrPath             : Union[str, IO[bytes]],
        writeIndex             : bool                      = False,
        clearIndexCache        : bool                      = False,
        indexFilePath          : Optional[str]             = None,
        indexFolders           : Optional[List[str]]       = None,
        encoding               : str                       = tarfile.ENCODING,
        verifyModificationTime : bool                      = False,
        printDebug             : int                       = 0,
        indexMinimumFileCount  : int                       = 1000,
        transform              : Optional[Tuple[str, str]] = None,
        **options
    ) -> None:
        # fmt: on
        if isinstance(fileOrPath, str):
            openedFile = True
            file: IO[bytes] = open(fileOrPath, 'rb')
        else:
            openedFile = False
            file = fileOrPath
            file.seek(0)

        offset = find_squashfs_offset(file)
        if offset < 0:
            if openedFile:
                file.close()
            raise ValueError("Not a valid SquashFS image!")

        # fmt: off
        self.rawFileObject          = file
        self.image                  = SquashFSImage(self.rawFileObject, offset=offset)
        self.archiveFilePath        = fileOrPath if isinstance(fileOrPath, str) else None
        self.encoding               = encoding
        self.verifyModificationTime = verifyModificationTime
        self.printDebug             = printDebug
        self.options                = options
        self.transformPattern       = transform
        # fmt: on

        self.transform = (
            (lambda x: re.sub(self.transformPattern[0], self.transformPattern[1], x))
            if isinstance(self.transformPattern, (tuple, list)) and len(self.transformPattern) == 2
            else (lambda x: x)
        )

        super().__init__(
            SQLiteIndex(
                indexFilePath,
                indexFolders=indexFolders,
                archiveFilePath=self.archiveFilePath,
                encoding=self.encoding,
                printDebug=self.printDebug,
                indexMinimumFileCount=indexMinimumFileCount,
                backendName='SquashFSMountSource',
            ),
            clearIndexCache=clearIndexCache,
            checkMetadata=self._check_metadata,
        )

        isFileObject = not isinstance(fileOrPath, str)

        if self.index.index_is_loaded():
            # self._load_or_store_compression_offsets()  # load
            self.index.reload_index_read_only()
        else:
            # Open new database when we didn't find an existing one.
            # Simply open in memory without an error even if writeIndex is True but when not indication
            # for a index file location has been given.
            if writeIndex and (indexFilePath or not isFileObject):
                self.index.open_writable()
            else:
                self.index.open_in_memory()

            self._create_index()
            # self._load_or_store_compression_offsets()  # store
            if self.index.index_is_loaded():
                self._store_metadata()
                self.index.reload_index_read_only()

    def _store_metadata(self) -> None:
        argumentsToSave = ['encoding', 'transformPattern']
        argumentsMetadata = json.dumps({argument: getattr(self, argument) for argument in argumentsToSave})
        self.index.store_metadata(argumentsMetadata, self.archiveFilePath)

    def _convert_to_row(self, inodeOffset: int, info: "PySquashfsImage.file.File") -> Tuple:  # type: ignore
        mode = info.mode
        linkname = ""
        if info.is_symlink:
            linkname = info.readlink()
            mode = mode | stat.S_IFLNK
        else:
            # Note that PySquashfsImage.file.Directory inherits from file.File, i.e., info can also be a directory.
            mode = mode | (stat.S_IFDIR if info.is_dir else stat.S_IFREG)

        path, name = SQLiteIndex.normpath(self.transform(info.path)).rsplit("/", 1)

        # Currently unused. Squashfs files are stored in multiple blocks, so a single offset is insufficient.
        dataOffset = 0

        # SquashFS also returns non-zero sizes for directory, FIFOs, symbolic links, and device files
        fileSize = info.size if info.is_file else 0

        # fmt: off
        fileInfo : Tuple = (
            path              ,  # 0  : path
            name              ,  # 1  : file name
            inodeOffset       ,  # 2  : header offset
            dataOffset        ,  # 3  : data offset
            fileSize          ,  # 4  : file size
            info.time         ,  # 5  : modification time
            mode              ,  # 6  : file mode / permissions
            0                 ,  # 7  : TAR file type. Currently unused. Overlaps with mode
            linkname          ,  # 8  : linkname
            0                 ,  # 9  : user ID
            0                 ,  # 10 : group ID
            False             ,  # 11 : is TAR (unused?)
            False             ,  # 12 : is sparse
            False             ,  # 13 : is generated (parent folder)
            0                 ,  # 14 : recursion depth
        )
        # fmt: on

        return fileInfo

    def _create_index(self) -> None:
        if self.printDebug >= 1:
            print(f"Creating offset dictionary for {self.archiveFilePath} ...")
        t0 = timer()

        self.index.ensure_intermediary_tables()

        # TODO Doing this in a chunked manner with generators would make it work better for large archives.
        fileInfos = []
        for inodeOffset, info in self.image:
            fileInfos.append(self._convert_to_row(inodeOffset, info))
        self.index.set_file_infos(fileInfos)

        # Resort by (path,name). This one-time resort is faster than resorting on each INSERT (cache spill)
        if self.printDebug >= 2:
            print("Resorting files by path ...")

        self.index.finalize()

        t1 = timer()
        if self.printDebug >= 1:
            print(f"Creating offset dictionary for {self.archiveFilePath} took {t1 - t0:.2f}s")

    def close(self) -> None:
        if hasattr(self, 'rawFileObject'):
            self.rawFileObject.close()

        # There is no "closed" method and it can only be closed once, else we get:
        # PySquashfsImage/__init__.py", line 131, in close
        #     self._fd.close()
        #     ^^^^^^^^^^^^^^
        # AttributeError: 'NoneType' object has no attribute 'close'
        with contextlib.suppress(AttributeError):
            self.image.close()  # pytype: disable=attribute-error

    @overrides(SQLiteIndexMountSource)
    def __exit__(self, exception_type, exception_value, exception_traceback) -> None:
        super().__exit__(exception_type, exception_value, exception_traceback)
        self.close()

    def __del__(self):
        self.close()
        if hasattr(super(), '__del__'):
            super().__del__()

    @overrides(MountSource)
    def open(self, fileInfo: FileInfo, buffering=-1) -> IO[bytes]:
        # The buffering is ignored for now because SquashFS has an inherent buffering based on the block size
        # configured in the SquashFS image. It probably makes no sense to reduce or increase that buffer size.
        # Decreasing may reduce memory usage, but with Python and other things, memory usage is not a priority
        # in ratarmount as long as it is bounded for very large archives.
        assert fileInfo.userdata
        extendedFileInfo = fileInfo.userdata[-1]
        assert isinstance(extendedFileInfo, SQLiteIndexedTarUserData)
        return self.image.open(self.image.read_inode(extendedFileInfo.offsetheader))

    @overrides(MountSource)
    def statfs(self) -> Dict[str, Any]:
        blockSize = 512
        with contextlib.suppress(Exception):
            blockSize = os.fstat(self.rawFileObject.fileno()).st_blksize

        blockSize = max(blockSize, self.image._sblk.block_size)
        return {
            'f_bsize': blockSize,
            'f_frsize': blockSize,
            'f_bfree': 0,
            'f_bavail': 0,
            'f_ffree': 0,
            'f_favail': 0,
            'f_namemax': 256,
        }

    def _check_metadata(self, metadata: Dict[str, Any]) -> None:
        """Raises an exception if the metadata mismatches so much that the index has to be treated as incompatible."""
        SQLiteIndex.check_archive_stats(self.archiveFilePath, metadata, self.verifyModificationTime)

        if 'arguments' in metadata:
            SQLiteIndex.check_metadata_arguments(
                json.loads(metadata['arguments']), self, argumentsToCheck=['encoding', 'transformPattern']
            )

        if 'backendName' not in metadata:
            self.index.try_to_open_first_file(lambda path: self.open(self.lookup(path)))



# pylint: disable=wrong-import-position
# pylint: disable=redefined-outer-name

import dataclasses
import io
import os
import stat
import sys
import tarfile
from pathlib import Path
from typing import Dict, List, Optional

import pytest

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from ratarmountcore.mountsource.compositing.subvolumes import SubvolumesMountSource  # noqa: E402
from ratarmountcore.mountsource.formats.folder import FolderMountSource  # noqa: E402
from ratarmountcore.mountsource.formats.tar import SQLiteIndexedTar  # noqa: E402
from ratarmountcore.mountsource.MountSource import MountSource  # noqa: E402


@dataclasses.dataclass
class SampleArchive:
    path: Path
    folders: List[str]
    files: Dict[str, bytes]


def _create_file(tarArchive, name, contents):
    tinfo = tarfile.TarInfo(name)
    tinfo.size = len(contents)
    tarArchive.addfile(tinfo, io.BytesIO(contents if isinstance(contents, bytes) else contents.encode()))


def _make_folder(tarArchive, name):
    tinfo = tarfile.TarInfo(name)
    tinfo.type = tarfile.DIRTYPE
    tarArchive.addfile(tinfo, io.BytesIO())


def _populate_folder(sampleArchive: SampleArchive):
    for folder in sampleArchive.folders:
        os.makedirs(os.path.join(sampleArchive.path, folder.strip('/')), exist_ok=True)
    for path, contents in sampleArchive.files.items():
        (sampleArchive.path / path.strip('/')).write_bytes(contents)


def _populate_tar(sampleArchive: SampleArchive):
    with tarfile.open(name=sampleArchive.path, mode="w:bz2") as tarFile:
        for folder in sampleArchive.folders:
            _make_folder(tarFile, folder)
        for path, contents in sampleArchive.files.items():
            _create_file(tarFile, path, contents)


@pytest.fixture(name="sample_folder_a")
def fixture_sample_folder_a(tmp_path):
    sampleArchive = SampleArchive(
        path=tmp_path / "folderA",
        folders=["subfolder"],
        files={"/subfolder/world": b"hello\n", "/ufo": b"iriya in folder 1\n"},
    )
    _populate_folder(sampleArchive)
    return sampleArchive


@pytest.fixture(name="sample_folder_b")
def fixture_sample_folder_b(tmp_path):
    sampleArchive = SampleArchive(
        path=tmp_path / "folderB",
        folders=["subfolder"],
        files={"/ufo": b"iriya\n"},
    )
    _populate_folder(sampleArchive)
    return sampleArchive


@pytest.fixture(name="sample_tar_a")
def fixture_sample_tar_a(tmp_path):
    sampleArchive = SampleArchive(
        path=tmp_path / "sampleA.tar",
        folders=["subfolder"],
        files={"/ufo": b"inside", "/README.md": b"readme inside", "/subfolder/world": b"HELLO"},
    )
    _populate_tar(sampleArchive)
    return sampleArchive


@pytest.fixture(name="sample_tar_b")
def fixture_sample_tar_b(tmp_path):
    sampleArchive = SampleArchive(
        path=tmp_path / "sampleB.tar",
        folders=["/src", "/dist", "/dist/a", "/dist/a/b"],
        files={"/README.md": b"hello world", "/src/test.sh": b"echo hi", "/dist/a/b/test2.sh": "echo two"},
    )
    _populate_tar(sampleArchive)
    return sampleArchive


class TestSubvolumesMountSource:
    @staticmethod
    def _check_file(mountSource: MountSource, path: str, version: int, contents: Optional[bytes] = None):
        fileInfo = mountSource.lookup(path, version)
        assert fileInfo, f"Path: {path}"

        if contents is None:
            assert stat.S_ISDIR(fileInfo.mode)
        else:
            assert not stat.S_ISDIR(fileInfo.mode)
            assert stat.S_ISREG(fileInfo.mode)

            # The MountSource interface only allows to open files in binary mode, which returns bytes not string.
            if isinstance(contents, str):
                contents = contents.encode()
            with mountSource.open(fileInfo) as file:
                assert file.read() == contents

    @staticmethod
    def test_unite_two_folders(sample_folder_a, sample_folder_b):
        union = SubvolumesMountSource(
            {"folderA": FolderMountSource(sample_folder_a.path), "folderB": FolderMountSource(sample_folder_b.path)}
        )

        assert union.lookup("folderC") is None

        # Check folders
        for path in sample_folder_a.folders:
            TestSubvolumesMountSource._check_file(union, "folderA/" + path, 0, None)
        for path in sample_folder_a.folders + sample_folder_b.folders:
            TestSubvolumesMountSource._check_file(union, "folderB/" + path, 0, None)

        # Check files
        for path, contents in sample_folder_a.files.items():
            TestSubvolumesMountSource._check_file(
                union, "folderA/" + path, 0 if path not in sample_folder_b.files else 1, contents
            )
        for path, contents in sample_folder_b.files.items():
            TestSubvolumesMountSource._check_file(union, "folderB/" + path, 0, contents)

    @staticmethod
    def test_unite_two_folders_and_update_one(sample_folder_a, sample_folder_b):
        folderA = FolderMountSource(sample_folder_a.path)
        folderB = FolderMountSource(sample_folder_b.path)
        volumes = {"folderA": folderA, "folderB": folderB}
        union = SubvolumesMountSource(volumes)

        contents = b"atarashii iriya\n"
        (sample_folder_a.path / "ufo2").write_bytes(contents)
        os.mkdir(os.path.join(sample_folder_a.path, "subfolder2"))
        (sample_folder_a.path / "subfolder2" / "world").write_bytes(contents)
        os.mkdir(os.path.join(sample_folder_a.path, "subfolder3"))
        (sample_folder_a.path / "subfolder3" / "world").write_bytes(contents)
        (sample_folder_a.path / "second-world").write_bytes(contents)

        TestSubvolumesMountSource._check_file(union, "folderA/ufo2", 0, contents)
        TestSubvolumesMountSource._check_file(union, "folderA/subfolder2", 0, None)
        TestSubvolumesMountSource._check_file(union, "folderA/subfolder2/world", 0, contents)
        TestSubvolumesMountSource._check_file(union, "folderA/subfolder3/world", 0, contents)
        TestSubvolumesMountSource._check_file(union, "folderA/second-world", 0, contents)

        # Test versions

        assert union.versions("/") == 1
        assert union.versions("") == 1
        assert union.versions("/NON_EXISTING") == 0
        assert union.versions("folderA/ufo") == 1
        assert union.versions("folderA/ufo2") == 1
        assert union.versions("folderB/ufo") == 1

        # Test list

        assert sorted(union.list("/").keys()) == ["folderA", "folderB"]
        assert sorted(union.list("").keys()) == ["folderA", "folderB"]
        assert sorted(union.list("/folderB").keys()) == ["subfolder", "ufo"]
        assert sorted(union.list("/folderB/subfolder").keys()) == []
        assert sorted(union.list("/folderA").keys()) == [
            "second-world",
            "subfolder",
            "subfolder2",
            "subfolder3",
            "ufo",
            "ufo2",
        ]

        # Test get_mount_source

        for folder in ["/", ""]:
            fileInfo = union.lookup(folder)
            assert fileInfo, f"Folder: {folder}"
            assert union.get_mount_source(fileInfo) == ("/", union, fileInfo)

        pathsBySubmount = {
            'folderA': ["folderA", "/folderA", "/folderA/ufo2", "folderA/subfolder2", "folderA/subfolder2/world"],
            'folderB': ["folderB", "/folderB", "/folderB/ufo", "folderB/subfolder"],
        }
        for volume, paths in pathsBySubmount.items():
            mountSource = volumes[volume]
            for path in paths:
                fileInfo = union.lookup(path)
                assert fileInfo, f"Path: {path}"
                result = union.get_mount_source(fileInfo)
                fileInfo.userdata.pop()
                assert result == ('/' + volume, mountSource, fileInfo)

    @staticmethod
    def test_unite_two_archives(sample_tar_a, sample_tar_b):
        union = SubvolumesMountSource(
            {"folderA": SQLiteIndexedTar(sample_tar_a.path), "folderB": SQLiteIndexedTar(sample_tar_b.path)}
        )

        # Check folders
        for path in sample_tar_a.folders:
            TestSubvolumesMountSource._check_file(union, "folderA/" + path, 0, None)
        for path in sample_tar_b.folders:
            TestSubvolumesMountSource._check_file(union, "folderB/" + path, 0, None)

        # Check files
        for path, contents in sample_tar_a.files.items():
            TestSubvolumesMountSource._check_file(
                union, "folderA/" + path, 0 if path not in sample_tar_b.files else 1, contents
            )
        for path, contents in sample_tar_b.files.items():
            TestSubvolumesMountSource._check_file(union, "folderB/" + path, 0, contents)

    @staticmethod
    def test_unite_folder_and_archive_and_update_folder(sample_tar_a, sample_folder_a):
        union = SubvolumesMountSource(
            {"tar": SQLiteIndexedTar(sample_tar_a.path), "folder": FolderMountSource(sample_folder_a.path)}
        )

        for path in sample_tar_a.folders:
            TestSubvolumesMountSource._check_file(union, "tar/" + path, 0, None)
        for path in sample_folder_a.folders:
            TestSubvolumesMountSource._check_file(union, "folder/" + path, 0, None)

        for path, contents in sample_tar_a.files.items():
            TestSubvolumesMountSource._check_file(
                union, "tar/" + path, 0 if path not in sample_folder_a.files else 1, contents
            )
        for path, contents in sample_folder_a.files.items():
            TestSubvolumesMountSource._check_file(union, "folder/" + path, 0, contents)

        # Create files inside the original folder
        contents = b"atarashii iriya\n"
        (sample_folder_a.path / "ufo2").write_bytes(contents)
        os.mkdir(os.path.join(sample_folder_a.path, "subfolder2"))
        (sample_folder_a.path / "subfolder2" / "world").write_bytes(contents)
        os.mkdir(os.path.join(sample_folder_a.path, "subfolder3"))
        (sample_folder_a.path / "subfolder3" / "world").write_bytes(contents)

        # Check for created files in the bind mount
        TestSubvolumesMountSource._check_file(union, "/folder/ufo2", 0, contents)
        TestSubvolumesMountSource._check_file(union, "/folder/subfolder2", 0, None)
        TestSubvolumesMountSource._check_file(union, "/folder/subfolder2/world", 0, contents)
        TestSubvolumesMountSource._check_file(union, "/folder/subfolder3/world", 0, contents)



# pylint: disable=wrong-import-position
# pylint: disable=redefined-outer-name

import dataclasses
import io
import os
import sys
import tarfile
from pathlib import Path
from typing import Dict, List

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import pytest  # noqa: E402
from ratarmountcore.mountsource.compositing.union import UnionMountSource  # noqa: E402
from ratarmountcore.mountsource.formats.folder import FolderMountSource  # noqa: E402
from ratarmountcore.mountsource.formats.tar import SQLiteIndexedTar  # noqa: E402


@dataclasses.dataclass
class SampleArchive:
    path: Path
    folders: List[str]
    files: Dict[str, bytes]


def _create_file(tarArchive, name, contents):
    tinfo = tarfile.TarInfo(name)
    tinfo.size = len(contents)
    tarArchive.addfile(tinfo, io.BytesIO(contents if isinstance(contents, bytes) else contents.encode()))


def _make_folder(tarArchive, name):
    tinfo = tarfile.TarInfo(name)
    tinfo.type = tarfile.DIRTYPE
    tarArchive.addfile(tinfo, io.BytesIO())


def _populate_folder(sampleArchive: SampleArchive):
    for folder in sampleArchive.folders:
        os.makedirs(os.path.join(sampleArchive.path, folder.strip('/')), exist_ok=True)
    for path, contents in sampleArchive.files.items():
        (sampleArchive.path / path.strip('/')).write_bytes(contents)


def _populate_tar(sampleArchive: SampleArchive):
    with tarfile.open(name=sampleArchive.path, mode="w:bz2") as tarFile:
        for folder in sampleArchive.folders:
            _make_folder(tarFile, folder)
        for path, contents in sampleArchive.files.items():
            _create_file(tarFile, path, contents)


@pytest.fixture(name="sample_folder_a")
def fixture_sample_folder_a(tmp_path):
    sampleArchive = SampleArchive(
        path=tmp_path / "folderA",
        folders=["subfolder"],
        files={"/subfolder/world": b"hello\n", "/ufo": b"iriya in folder 1\n"},
    )
    _populate_folder(sampleArchive)
    return sampleArchive


@pytest.fixture(name="sample_folder_b")
def fixture_sample_folder_b(tmp_path):
    sampleArchive = SampleArchive(
        path=tmp_path / "folderB",
        folders=["subfolder"],
        files={"/ufo": b"iriya\n"},
    )
    _populate_folder(sampleArchive)
    return sampleArchive


@pytest.fixture(name="sample_tar_a")
def fixture_sample_tar_a(tmp_path):
    sampleArchive = SampleArchive(
        path=tmp_path / "sampleA.tar",
        folders=["subfolder"],
        files={"/ufo": b"inside", "/README.md": b"readme inside", "/subfolder/world": b"HELLO"},
    )
    _populate_tar(sampleArchive)
    return sampleArchive


@pytest.fixture(name="sample_tar_b")
def fixture_sample_tar_b(tmp_path):
    sampleArchive = SampleArchive(
        path=tmp_path / "sampleB.tar",
        folders=["/src", "/dist", "/dist/a", "/dist/a/b"],
        files={"/README.md": b"hello world", "/src/test.sh": b"echo hi", "/dist/a/b/test2.sh": "echo two"},
    )
    _populate_tar(sampleArchive)
    return sampleArchive


class TestUnionMountSource:
    @staticmethod
    def _check_file(mountSource, path, version, contents=None):
        fileInfo = mountSource.lookup(path, version)
        assert fileInfo, f"Path: {path}"
        if contents is not None:
            # The MountSource interface only allows to open files in binary mode, which returns bytes not string.
            if isinstance(contents, str):
                contents = contents.encode()
            with mountSource.open(fileInfo) as file:
                assert file.read() == contents

    @staticmethod
    def test_unite_one_archive(sample_tar_a):
        union = UnionMountSource([SQLiteIndexedTar(sample_tar_a.path)])

        assert union.is_immutable()

        for path in sample_tar_a.folders:
            TestUnionMountSource._check_file(union, path, 0, None)
        for path, contents in sample_tar_a.files.items():
            TestUnionMountSource._check_file(union, path, 0, contents)

    @staticmethod
    def test_unite_two_folders(sample_folder_a, sample_folder_b):
        union = UnionMountSource([FolderMountSource(sample_folder_a.path), FolderMountSource(sample_folder_b.path)])
        for path in sample_folder_a.folders + sample_folder_b.folders:
            TestUnionMountSource._check_file(union, path, 0, None)
        for path, contents in sample_folder_b.files.items():
            TestUnionMountSource._check_file(union, path, 0, contents)
        for path, contents in sample_folder_a.files.items():
            TestUnionMountSource._check_file(union, path, 0 if path not in sample_folder_b.files else 1, contents)

    @staticmethod
    def test_unite_two_folders_and_update_one(sample_folder_a, sample_folder_b):
        union = UnionMountSource([FolderMountSource(sample_folder_a.path), FolderMountSource(sample_folder_b.path)])

        contents = b"atarashii iriya\n"
        (sample_folder_a.path / "ufo2").write_bytes(contents)
        os.mkdir(os.path.join(sample_folder_a.path, "subfolder2"))
        (sample_folder_a.path / "subfolder2" / "world").write_bytes(contents)
        os.mkdir(os.path.join(sample_folder_a.path, "subfolder3"))
        (sample_folder_a.path / "subfolder3" / "world").write_bytes(contents)
        (sample_folder_a.path / "second-world").write_bytes(contents)

        TestUnionMountSource._check_file(union, "/ufo2", 0, contents)
        TestUnionMountSource._check_file(union, "/subfolder2", 0, None)
        TestUnionMountSource._check_file(union, "/subfolder2/world", 0, contents)
        TestUnionMountSource._check_file(union, "/subfolder3/world", 0, contents)
        TestUnionMountSource._check_file(union, "/second-world", 0, contents)

    @staticmethod
    def test_unite_two_archives(sample_tar_a, sample_tar_b):
        union = UnionMountSource([SQLiteIndexedTar(sample_tar_a.path), SQLiteIndexedTar(sample_tar_b.path)])
        for path in sample_tar_a.folders + sample_tar_b.folders:
            TestUnionMountSource._check_file(union, path, 0, None)
        for path, contents in sample_tar_b.files.items():
            TestUnionMountSource._check_file(union, path, 0, contents)
        for path, contents in sample_tar_a.files.items():
            TestUnionMountSource._check_file(union, path, 0 if path not in sample_tar_b.files else 1, contents)

    @staticmethod
    def test_unite_folder_and_archive_and_update_folder(sample_tar_a, sample_folder_a):
        archive = SQLiteIndexedTar(sample_tar_a.path)
        folder = FolderMountSource(sample_folder_a.path)
        union = UnionMountSource([archive, folder])
        for path in sample_tar_a.folders + sample_folder_a.folders:
            TestUnionMountSource._check_file(union, path, 0, None)
        for path, contents in sample_folder_a.files.items():
            TestUnionMountSource._check_file(union, path, 0, contents)
        for path, contents in sample_tar_a.files.items():
            TestUnionMountSource._check_file(union, path, 0 if path not in sample_folder_a.files else 1, contents)

        assert not union.is_immutable()

        contents = b"atarashii iriya\n"
        (sample_folder_a.path / "ufo2").write_bytes(contents)
        os.mkdir(os.path.join(sample_folder_a.path, "subfolder2"))
        (sample_folder_a.path / "subfolder2" / "world").write_bytes(contents)
        os.mkdir(os.path.join(sample_folder_a.path, "subfolder3"))
        (sample_folder_a.path / "subfolder3" / "world").write_bytes(contents)

        TestUnionMountSource._check_file(union, "/ufo2", 0, contents)
        TestUnionMountSource._check_file(union, "/subfolder2", 0, None)
        TestUnionMountSource._check_file(union, "/subfolder2/world", 0, contents)
        TestUnionMountSource._check_file(union, "/subfolder3/world", 0, contents)

        for path in ["/", ""]:
            fileInfo = union.lookup(path)
            assert fileInfo, f"Path: {path}"
            assert union.get_mount_source(fileInfo) == ("/", union, fileInfo)

        # Test versions

        assert union.versions("/") == 2
        assert union.versions("") == 2
        assert union.versions("/NON_EXISTING") == 0
        assert union.versions("subfolder") == 2
        assert union.versions("/subfolder") == 2
        assert union.versions("ufo") == 2
        assert union.versions("/ufo") == 2
        assert union.versions("ufo2") == 1
        assert union.versions("/ufo2") == 1

        # Test list

        rootFiles = ["README.md", "subfolder", "subfolder2", "subfolder3", "ufo", "ufo2"]
        assert sorted(union.list("/").keys()) == rootFiles
        assert sorted(union.list("").keys()) == rootFiles
        assert sorted(union.list("/subfolder").keys()) == ["world"]
        assert sorted(union.list("/subfolder2").keys()) == ["world"]
        assert sorted(union.list("/subfolder3").keys()) == ["world"]

        # Test files in folder, which has higher precedence than the archive.
        for path in ["/ufo", "/ufo2", "ufo2", "/subfolder2", "subfolder2", "/subfolder2/world"]:
            fileInfo = union.lookup(path)
            assert fileInfo, f"Path: {path}"
            result = union.get_mount_source(fileInfo)
            fileInfo.userdata.pop()
            assert result == ("/", folder, fileInfo)

        # Test files exclusively in archive.
        for path in ["/README.md", "README.md"]:
            fileInfo = union.lookup(path)
            assert fileInfo, f"Path: {path}"
            result = union.get_mount_source(fileInfo)
            fileInfo.userdata.pop()
            assert result == ("/", archive, fileInfo)

        # Test files that exist in both.
        for path in ["/ufo", "ufo", "subfolder/world", "/subfolder/world"]:
            fileInfo = union.lookup(path, fileVersion=1)
            assert fileInfo, f"Path: {path}"
            result = union.get_mount_source(fileInfo)
            fileInfo.userdata.pop()
            assert result == ("/", archive, fileInfo)



# pylint: disable=wrong-import-order
# pylint: disable=wrong-import-position
# pylint: disable=protected-access

import bz2
import hashlib
import io
import os
import stat
import sys

import pytest

try:
    import ext4
except ImportError:
    ext4 = None  # type:ignore

from helpers import copy_test_file

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from ratarmountcore.mountsource.formats.ext4 import EXT4MountSource  # noqa: E402


class TestEXT4MountSource:
    @staticmethod
    @pytest.mark.parametrize('path', ['nested-tar-1M.ext4.bz2', 'nested-tar-10M.ext4.bz2'])
    def test_password(path):
        if not ext4:
            return
        with copy_test_file(path) as tmpPath, bz2.open(tmpPath, 'rb') as bz2File:
            tmpFileObject = io.BytesIO()
            tmpFileObject.write(bz2File.read())
            tmpFileObject.seek(0)

            mountSource = EXT4MountSource(tmpFileObject)
            for folder in ['/', '/foo', '/foo/fighter']:
                fileInfo = mountSource.lookup(folder)
                assert fileInfo
                assert stat.S_ISDIR(fileInfo.mode)

                assert mountSource.versions(folder) == 1
                assert mountSource.list(folder)

            for filePath in ['/foo/fighter/ufo', '/foo/lighter.tar']:
                fileInfo = mountSource.lookup(filePath)
                assert fileInfo
                assert not stat.S_ISDIR(fileInfo.mode)

                assert mountSource.versions(filePath) == 1
                assert not mountSource.list(filePath)

            with mountSource.open(mountSource.lookup('/foo/fighter/ufo')) as file:
                assert file.read() == b'iriya\n'

            with mountSource.open(mountSource.lookup('/foo/lighter.tar')) as file:
                assert hashlib.md5(file.read()).hexdigest() == "2a06cc391128d74e685a6cb7cfe9f94d"

    # TODO Does not use SQLiteIndex backend because it is already usably indexed and it would seem redundant.
    #      Therefore 'transform' does not work.



import os
import re
import stat
import traceback
from dataclasses import dataclass
from typing import IO, Any, Dict, Iterable, List, Optional, Tuple, Union

from ratarmountcore.compressions import strip_suffix_from_archive
from ratarmountcore.mountsource import FileInfo, MountSource, merge_statfs
from ratarmountcore.mountsource.factory import open_mount_source
from ratarmountcore.mountsource.formats.folder import FolderMountSource
from ratarmountcore.mountsource.formats.tar import SQLiteIndexedTar, SQLiteIndexedTarUserData
from ratarmountcore.utils import RatarmountError, determine_recursion_depth, overrides


class AutoMountLayer(MountSource):
    """
    This mount source takes another mount source and automatically shows the contents of files which are archives.
    The detailed behavior can be controlled using options.
    """

    __slots__ = ('mounted', 'options')

    @dataclass
    class MountInfo:
        mountSource: MountSource
        rootFileInfo: FileInfo
        recursionDepth: int

    def __init__(self, mountSource: MountSource, **options) -> None:
        if isinstance(mountSource, AutoMountLayer):
            # Stacking can make sense and should work when they have non-overlapping(!) recursion depths.
            raise RatarmountError("AutoMountLayer must not be stacked directly onto each other.")

        self.options = options
        self.maxRecursionDepth: int = determine_recursion_depth(**options)
        self.lazyMounting: bool = self.options.get('lazyMounting', False)
        self.printDebug = int(options.get("printDebug", 0)) if isinstance(options.get("printDebug", 0), int) else 0

        rootFileInfo = mountSource.lookup('/')
        assert rootFileInfo
        rootFileInfo.userdata.append('/')

        # Mount points are specified without trailing slash and with leading slash
        # representing root of this mount source.
        # Disable false positive introduced when updating pylint from 2.6 to 2.12.
        # It now thinks that the assignment is to AutoMountLayer instead of self.mounted.
        # pylint: disable=used-before-assignment
        self.mounted: Dict[str, AutoMountLayer.MountInfo] = {
            '/': AutoMountLayer.MountInfo(mountSource, rootFileInfo, 0)
        }

        if self.lazyMounting:
            return

        # Go over all files and mount archives and even archives in those archives
        foldersToWalk = ['/']
        while foldersToWalk:
            newFoldersToWalk = []
            for folder in foldersToWalk:
                if self.get_recursion_depth(folder) > self.maxRecursionDepth:
                    continue

                fileNames = self.list(folder)
                if not fileNames:
                    continue

                for fileName in fileNames:
                    filePath = os.path.join(folder, fileName)
                    if self.is_dir(filePath):
                        newFoldersToWalk.append(filePath)
                    else:
                        mountPoint = self._try_to_mount_file(filePath)
                        if mountPoint:
                            newFoldersToWalk.append(mountPoint)

            foldersToWalk = newFoldersToWalk

    def _simply_find_mounted(self, path: str) -> Tuple[str, str]:
        """See _find_mounted. This is split off to avoid convoluted recursions during lazy mounting."""

        leftPart = path
        rightParts: List[str] = []
        while '/' in leftPart:
            if leftPart in self.mounted:
                return leftPart, '/' + '/'.join(rightParts)

            parts = leftPart.rsplit('/', 1)
            leftPart = parts[0]
            rightParts.insert(0, parts[1])

        assert '/' in self.mounted
        return '/', path

    def get_recursion_depth(self, path: str) -> int:
        mountPoint, pathInMountPoint = self._simply_find_mounted(path)
        mountInfo = self.mounted[mountPoint]
        fileInfo = mountInfo.mountSource.lookup(pathInMountPoint)

        # +1 because, by definition each mount source adds one recursion.
        # There can be no passthrough MountSource here because they are only created on archives..
        return mountInfo.recursionDepth + (
            sum(
                userdata.recursiondepth
                for userdata in fileInfo.userdata
                if isinstance(userdata, SQLiteIndexedTarUserData)
            )
            + 1
            if fileInfo
            else 0
        )

    def _try_to_mount_file(self, path: str) -> Optional[str]:
        """
        Returns the mount point path if it has been successfully mounted.
        path: Path inside this mount source. May include recursively mounted mount points.
              Should contain a leading slash.
        """

        # For better performance, only look at the suffix not at the magic bytes.
        strippedFilePath = strip_suffix_from_archive(path)
        if strippedFilePath == path:
            return None

        recursionDepth = self.get_recursion_depth(path)
        if recursionDepth > self.maxRecursionDepth:
            return None

        mountPoint = strippedFilePath if self.options.get('stripRecursiveTarExtension', False) else path
        # https://unix.stackexchange.com/questions/655155/how-to-repeatedly-unpack-tar-gz-files-that-are-within-the-tar-gz-itself
        if 'transformRecursiveMountPoint' in self.options:
            pattern = self.options['transformRecursiveMountPoint']
            if isinstance(pattern, (tuple, list)) and len(pattern) == 2:
                mountPoint = '/' + re.sub(pattern[0], pattern[1], mountPoint).lstrip('/')

        if mountPoint in self.mounted:
            return None

        # Use _simply_find_mounted instead of _find_mounted or self.open to avoid recursions caused by lazy mounting!
        parentMountPoint, pathInsideParentMountPoint = self._simply_find_mounted(path)
        parentMountInfo = self.mounted[parentMountPoint]
        parentMountSource = parentMountInfo.mountSource

        archiveFileInfo = parentMountSource.lookup(pathInsideParentMountPoint)
        if archiveFileInfo is None:
            return None

        # Do not mount uncompressed TARs inside SQLiteIndexedTar when they already were mounted recursively!
        mountSourceResult = parentMountSource.get_mount_source(archiveFileInfo)
        if mountSourceResult:
            realMountSource = mountSourceResult[1]
            if (
                isinstance(realMountSource, SQLiteIndexedTar)
                and realMountSource.mountRecursively
                and archiveFileInfo.userdata
            ):
                indexedTarData = archiveFileInfo.userdata[0]
                if isinstance(indexedTarData, SQLiteIndexedTarUserData) and indexedTarData.istar:
                    return None

        recursionDepth = self.get_recursion_depth(path)

        try:
            options = self.options.copy()
            options['recursionDepth'] = max(0, self.maxRecursionDepth - recursionDepth)

            _, deepestMountSource, deepestFileInfo = parentMountSource.get_mount_source(archiveFileInfo)
            if isinstance(deepestMountSource, FolderMountSource):
                # Open from file path on host file system in order to write out TAR index files.
                # Care has to be taken if a folder is bind mounted onto itself because then it can happen that
                # the file open triggers a recursive FUSE call, which then hangs up everything.
                mountSource = open_mount_source(deepestMountSource.get_file_path(deepestFileInfo), **options)
            else:
                # This will fail with StenciledFile objects as returned by SQLiteIndexedTar mount sources and when
                # given to backends like indexed_zstd, which do expect the file object to have a valid fileno.
                mountSource = open_mount_source(
                    parentMountSource.open(archiveFileInfo),
                    tarFileName=pathInsideParentMountPoint.rsplit('/', 1)[-1],
                    **options,
                )
        except Exception as e:
            print("[Warning] Mounting of '" + path + "' failed because of:", e)
            if self.printDebug >= 3:
                traceback.print_exc()
            print()
            return None

        rootFileInfo = archiveFileInfo.clone()
        rootFileInfo.mode = (rootFileInfo.mode & 0o777) | stat.S_IFDIR
        rootFileInfo.linkname = ""
        rootFileInfo.userdata = [mountPoint]
        mountInfo = AutoMountLayer.MountInfo(mountSource, rootFileInfo, recursionDepth)

        # TODO What if the mount point already exists, e.g., because stripRecursiveTarExtension is true and there
        #      are multiple archives with the same name but different extensions?
        self.mounted[mountPoint] = mountInfo
        if self.printDebug >= 2:
            print("Recursively mounted:", mountPoint)
            print()

        return mountPoint

    def _find_mounted(self, path: str) -> Tuple[str, str]:
        """
        Returns the mount point, which can be found in self.mounted, and the rest of the path.
        Basically, it splits path at the appropriate mount point boundary.
        Because of the recursive mounting, there might be multiple mount points fitting the path.
        The longest, i.e., the deepest mount point will be returned.
        """

        if self.maxRecursionDepth and self.lazyMounting:
            subPath = "/"
            # First go from higher paths to deeper ones and try to mount all parent archives lazily.
            for part in path.lstrip('/').split('/'):
                subPath = os.path.join(subPath, part)

                if self.get_recursion_depth(subPath) > self.maxRecursionDepth:
                    break

                if subPath not in self.mounted:
                    self._try_to_mount_file(subPath)

        return self._simply_find_mounted(path)

    @overrides(MountSource)
    def is_immutable(self) -> bool:
        return self.mounted['/'].mountSource.is_immutable()

    @overrides(MountSource)
    def lookup(self, path: str, fileVersion: int = 0) -> Optional[FileInfo]:
        """
        Return file info for given path. Note that all returned file infos contain MountInfo
        or a file path string at the back of FileInfo.userdata.
        """
        # TODO: Add support for the .versions API in order to access the underlying TARs if stripRecursiveTarExtension
        #       is false? Then again, SQLiteIndexedTar is not able to do this either, so it might be inconsistent.

        # It might be arguably that we could simply let the mount source handle returning file infos for the root
        # directory but only we know the permissions of the parent folder and can apply them to the root directory.
        mountPoint, pathInMountPoint = self._find_mounted(path)
        mountInfo = self.mounted[mountPoint]

        originalFileVersions = 0
        if mountPoint != '/' and pathInMountPoint == '/':
            originalFileVersions = self.mounted['/'].mountSource.versions(path)

        def normalize_file_version(version, versions):
            return ((version - 1) % versions + 1) % versions if versions > 1 else version

        # fileVersion=0 is the most recent. Version 1..fileVersions number from the first occurrence / oldest
        # version to the most recent, i.e., fileVersion = 0 is equivalent to fileVersion = fileVersions.
        fileVersions = self.versions(path)
        fileVersion = normalize_file_version(fileVersion, fileVersions)
        if fileVersion == 0 and pathInMountPoint == '/':
            return mountInfo.rootFileInfo.clone()

        if fileVersions <= 1 or pathInMountPoint != '/' or fileVersion == 0 or fileVersion > originalFileVersions:
            fileInfo = mountInfo.mountSource.lookup(pathInMountPoint, fileVersion - originalFileVersions)
            if fileInfo:
                fileInfo.userdata.append(mountPoint)
            return fileInfo

        # We are here if: fileVersions > 1 and 0 < fileVersion <= originalFileVersions and pathInMountPoint == '/'
        fileInfo = self.mounted['/'].mountSource.lookup(path, fileVersion % originalFileVersions)
        if fileInfo:
            fileInfo.userdata.append('/')
        return fileInfo

    def _append_mount_points(self, path: str, files, onlyMode: bool):
        if not files:
            return None

        if not isinstance(files, dict):
            files = set(files)

        # Check whether we need to add recursive mount points to this directory listing
        # The outer if is only a performance optimization. In general, it should be possible to remove it.
        # In case that 'files' also contains stat results, we might have to overwrite some results with
        # the results for a mount point!
        if (
            not isinstance(files, set)
            or self.options.get('stripRecursiveTarExtension', False)
            or self.options.get('transformRecursiveMountPoint', False)
        ):
            for mountPoint, mountInfo in self.mounted.items():
                folder, folderName = os.path.split(mountPoint)
                # This also potentially updates a archive file stat result with a stat result for a folder type!
                if folder == path and folderName and folderName:
                    if isinstance(files, set):
                        files.add(folderName)
                    else:
                        files.update(
                            {folderName: mountInfo.rootFileInfo.mode if onlyMode else mountInfo.rootFileInfo.clone()}
                        )

        return files

    @overrides(MountSource)
    def list(self, path: str) -> Optional[Union[Iterable[str], Dict[str, FileInfo]]]:
        mountPoint, pathInMountPoint = self._find_mounted(path)
        return self._append_mount_points(
            path, self.mounted[mountPoint].mountSource.list(pathInMountPoint), onlyMode=False
        )

    @overrides(MountSource)
    def list_mode(self, path: str) -> Optional[Union[Iterable[str], Dict[str, int]]]:
        mountPoint, pathInMountPoint = self._find_mounted(path)
        return self._append_mount_points(
            path, self.mounted[mountPoint].mountSource.list_mode(pathInMountPoint), onlyMode=True
        )

    @overrides(MountSource)
    def versions(self, path: str) -> int:
        mountPoint, pathInMountPoint = self._find_mounted(path)
        fileVersions = self.mounted[mountPoint].mountSource.versions(pathInMountPoint)
        if mountPoint != '/' and pathInMountPoint == '/':
            fileVersions += self.mounted['/'].mountSource.versions(path)
        return fileVersions

    @overrides(MountSource)
    def open(self, fileInfo: FileInfo, buffering=-1) -> IO[bytes]:
        _, mountSource, sourceFileInfo = self.get_mount_source(fileInfo)
        return mountSource.open(sourceFileInfo, buffering=buffering)

    @overrides(MountSource)
    def read(self, fileInfo: FileInfo, size: int, offset: int) -> bytes:
        _, mountSource, sourceFileInfo = self.get_mount_source(fileInfo)
        return mountSource.read(sourceFileInfo, size, offset)

    @overrides(MountSource)
    def list_xattr(self, fileInfo: FileInfo) -> List[str]:
        mountPoint = fileInfo.userdata[-1]
        assert isinstance(mountPoint, str)
        if fileInfo == self.mounted[mountPoint].rootFileInfo:
            return []

        _, mountSource, sourceFileInfo = self.get_mount_source(fileInfo)
        return mountSource.list_xattr(sourceFileInfo)

    @overrides(MountSource)
    def get_xattr(self, fileInfo: FileInfo, key: str) -> Optional[bytes]:
        mountPoint = fileInfo.userdata[-1]
        assert isinstance(mountPoint, str)
        if fileInfo == self.mounted[mountPoint].rootFileInfo:
            return None

        _, mountSource, sourceFileInfo = self.get_mount_source(fileInfo)
        return mountSource.get_xattr(sourceFileInfo, key)

    @overrides(MountSource)
    def get_mount_source(self, fileInfo: FileInfo) -> Tuple[str, MountSource, FileInfo]:
        mountPoint = fileInfo.userdata[-1]
        assert isinstance(mountPoint, str)
        mountSource = self.mounted[mountPoint].mountSource

        sourceFileInfo = fileInfo.clone()
        sourceFileInfo.userdata.pop()

        deeperMountPoint, deeperMountSource, deeperFileInfo = mountSource.get_mount_source(sourceFileInfo)
        return os.path.join(mountPoint, deeperMountPoint.lstrip('/')), deeperMountSource, deeperFileInfo

    @overrides(MountSource)
    def statfs(self) -> Dict[str, Any]:
        return merge_statfs(
            [mountInfo.mountSource.statfs() for _, mountInfo in self.mounted.items()], printDebug=self.printDebug
        )

    @overrides(MountSource)
    def __exit__(self, exception_type, exception_value, exception_traceback):
        for mountInfo in self.mounted.values():
            mountInfo.mountSource.__exit__(exception_type, exception_value, exception_traceback)

    def join_threads(self):
        for mountInfo in self.mounted.values():
            if hasattr(mountInfo.mountSource, 'join_threads'):
                mountInfo.mountSource.join_threads()
